#!/usr/bin/env python
#####################################################################################
#
# Library for configuring service discovery
# Joachim Flammer <Joachim.Flammer@cern.ch>
# $Id: serviceDiscovery.py,v 1.2.2.7 2005/09/14 08:37:59 flammer Exp $
#
#####################################################################################
#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://eu-egee.org/partners/ for details on the copyright holders.
# For license conditions see the license file or http://eu-egee.org/license.html
#
#####################################################################################

import os
import xml.dom.minidom
from xml.dom.minidom import getDOMImplementation
import _xmlplus.xpath as xpath
    
from gLiteInstallerLib import gLib
from xml.dom.ext import PrettyPrint

class ServiceDiscovery:
    
    #################################################################
    # init
    #
    # init class
    #
    # return 0 ok
    #     1 error
    #     
    #################################################################
    def __init__(self):

        self.glib = gLib()
        self.verbose = 0        
        
        return

    #################################################################
    # configure
    #
    # configure service discovery
    #
    # return 0 ok
    #     1 error
    #     
    #################################################################
    def configure(self,params,glib):

        print "Configuring Service Discovery ... "        
            
        # ------------------------------------------------------------        
        # get generic service discovery parameters
        # ------------------------------------------------------------

        # service discovery type
        if params.has_key('service-discovery.type'):
            service_discovery_type = params['service-discovery.type']
        else:
            glib.printFailedMessage()
            glib.printErrorMessage("Parameter \"service-discovery.type\" not found.")
            return 1
                
        # service discovery default site
        if params.has_key('service-discovery.site'):
            service_discovery_site = params['service-discovery.site']
        else:
            service_discovery_site = ""

        # service discovery default site
        if params.has_key('service-discovery.vo'):
            service_discovery_vo = params['service-discovery.vo']
        else:
            service_discovery_vo = ""
        
        # ------------------------------------------------------------        
        # set environment parameters
        # ------------------------------------------------------------
        
        first = 1
        service_discovery_plugin = ""
        for type in service_discovery_type:
            if first:
                first = 0
            else:
                service_discovery_plugin = service_discovery_plugin + ","
            service_discovery_plugin = service_discovery_plugin + type    
        glib.export('GLITE_SD_PLUGIN', service_discovery_plugin)
        
        if service_discovery_site != "":            
            glib.export('GLITE_SD_SITE', service_discovery_site)
        else:
            glib.unexport('GLITE_SD_SITE')
                    
        if service_discovery_vo != "":            
            glib.export('GLITE_SD_VO', service_discovery_vo)
        else:
        	glib.unexport('GLITE_SD_VO')
        
        # ------------------------------------------------------------        
        # configure individual service discovery types
        # ------------------------------------------------------------
        
        selected = 0
        errorcode = 0
        
        for type in service_discovery_type:
            
            # configure file based service discovery
            if type.lower() == "file":
                errorcode += self.configureFileBasedSD(glib)
                selected += 1
            else:
                # configure bdii based service discovery
                if type.lower() == "bdii":
                    errorcode += self.configureBdiiBasedSD(glib)
                    selected += 1
                else:
                    # configure R-GMA based service discovery                    
                    if type.lower() == "rgma":
                        errorcode += self.configureRgmaBasedSD(glib)
                        selected += 1
            
        if errorcode:
            print "Configuring Service Discovery                             ", 
            glib.printFailedMessage()
            return 1
        
        print "Configuring Service Discovery                             ",
        glib.printOkMessage()    
        
        return 0


    #################################################################
    # configureBdiiBasedSD
    #
    # configure bdii based service discovery
    #
    # return 0 ok
    #         1 error
    #     
    #################################################################
    def configureBdiiBasedSD(self,glib):
        
        print "Configuring BDII based Service Discovery                  ",
        
        params = glib.params
        
        # service discovery default site
        if params.has_key('service-discovery.bdii.provider'):
            service_discovery_bdii_provider = params['service-discovery.bdii.provider']
        else:
            glib.printFailedMessage()
            glib.printErrorMessage("Parameter \"service-discovery.bdii.provider\" not found.")
            return 1
            
        glib.export('LCG_GFAL_INFOSYS', service_discovery_bdii_provider)
        
        glib.printOkMessage()
        
        return 0
        

    #################################################################
    # configureRgmaBasedSD
    #
    # configure rgma based service discovery
    #
    # return 0 ok
    #         1 error
    #     
    #################################################################
    def configureRgmaBasedSD(self,glib):
                
        print "Configuring R-GMA based Service Discovery ...             "
            
        params = glib.params
        
        # ------------------------------------------------------------        
        # check if R-GMA client is installed
        # ------------------------------------------------------------        
        if not os.path.exists("%s/glite-rgma-client-config.py" % glib.getScriptPath()):
            glib.printErrorMessage("R-GMA client is needed to run R-GMA based service discovery.")
            glib.printErrorMessage("R-GMA client is not installed.")
            glib.printErrorMessage("Please install the R-GMA client and run the configuration again.")
            print "Configuring R-GMA based Service Discovery                 ",
            glib.printFailedMessage()
            return 1
        
        # ------------------------------------------------------------        
        # configure R-GMA client
        # ------------------------------------------------------------
        if self.verbose:
            print "Configuring R-GMA client ...                              "    
        
        cmd = "python %s/glite-rgma-client-config.py --configure --subservice" % glib.getScriptPath()
        if os.system(cmd):
            print "Configuring R-GMA client                                  ",
            glib.printFailedMessage()
            print "Configuring R-GMA based Service Discovery                 ",
            glib.printFailedMessage()
            return 1

        print "Configuring R-GMA client                                  ",
        glib.printOkMessage()

        print "Configuring R-GMA based Service Discovery                 ",
        glib.printOkMessage()
                
        return 0
        
        
    #################################################################
    # configureFileBasedSD
    #
    # configure file based service discovery
    #
    # return 0 ok
    #         1 error
    #     
    #################################################################
    def configureFileBasedSD(self,glib):
        
        if self.verbose:
            print "Configuring file based Service Discovery ... "    
        else:
            print "Configuring file based Service Discovery                  ",
        
        # service discovery file 
        self.serviceDiscoveryFile = '%s/etc/services.xml' % os.environ['GLITE_LOCATION']
        
        if os.path.exists(self.serviceDiscoveryFile):
            # load existing document
            if self.verbose:
                print "Loading existing service discovery description from file  ",        
            self.doc = xml.dom.minidom.parse(self.serviceDiscoveryFile)
            if self.verbose:
                glib.printOkMessage()
        else:
            # create new document
            if self.verbose:
                print "Creating new service discovery description file           ",
            impl = getDOMImplementation()
            self.doc = impl.createDocument(None, "services", None)
            if self.verbose:
                glib.printOkMessage()
        
        # get instances for service discovery
        instances = glib.getInstances('service-discovery.file')
        if len(instances) > 0:
            for instanceName in instances:
                                                        
                # get parameter for one instance
                instanceParams = {}
                try:
                    instanceParams = glib.getInstanceParams(instanceName)
                except ValueError, msg:
                    if not self.verbose:
                        print "\n"
                    glib.printErrorMessage("%s - Could not get parameters. Please check configuration files." % msg)
                    return 1
                except LookupError, msg:
                    print msg
                        
                # check existance of all necessary parameters
                
                # service_name => required
                if instanceParams.has_key('service-discovery.file.service_name'):
                    serviceName = instanceParams['service-discovery.file.service_name']
                else:
                    if not self.verbose:
                        print "\n"
                    glib.printErrorMessage("Parameter \"service-discovery.file.service_name\" not found.")
                    return 1
                
                # service_type => required
                if instanceParams.has_key('service-discovery.file.service_type'):
                    serviceType = instanceParams['service-discovery.file.service_type']
                else:
                    if not self.verbose:
                        print "\n"                        
                    glib.printErrorMessage("Parameter \"service-discovery.file.service_type\" not found.")
                    return 1    
                
                # service_version => required
                if instanceParams.has_key('service-discovery.file.service_version'):
                    serviceVersion = instanceParams['service-discovery.file.service_version']
                else:
                    if not self.verbose:
                        print "\n"                        
                    glib.printErrorMessage("Parameter \"service-discovery.file.service_version\" not found.")
                    return 1
                
                # service url endpoint => required
                if instanceParams.has_key('service-discovery.file.url_endpoint'):
                    serviceEndpoint = instanceParams['service-discovery.file.url_endpoint']
                else:
                    if not self.verbose:
                        print "\n"                        
                    glib.printErrorMessage("Parameter \"service-discovery.file.url_endpoint\" not found.")
                    return 1
                    
                # site => optional
                if instanceParams.has_key('service-discovery.file.site'):
                    serviceSite = instanceParams['service-discovery.file.site']
                else:
                    serviceSite = ""

                # vo list => optional
                serviceVos = []
                if instanceParams.has_key('service-discovery.file.vo'):
                    if len(instanceParams['service-discovery.file.vo']) > 0:
                        for item in instanceParams['service-discovery.file.vo']:
                            if item != "":
                                serviceVos.append(item) 
    
                # wsdl => optional
                if instanceParams.has_key('service-discovery.file.wsdl'):
                    serviceWsdl = instanceParams['service-discovery.file.wsdl']
                else:
                    serviceWsdl = ""

                # administration => optional
                if instanceParams.has_key('service-discovery.file.administration'):
                    serviceAdministration = instanceParams['service-discovery.file.administration']
                else:
                    serviceAdministration = ""

                # param list => optional
                serviceParams = []
                if instanceParams.has_key('service-discovery.file.param'):
                    if len(instanceParams['service-discovery.file.param']) > 0:
                        for item in instanceParams['service-discovery.file.param']:
                            if item.find('=') != -1:
                                serviceParams.append(item)
                            else:
                                if item !="":
                                    glib.printErrorMessage("Parameter 'service-discovery.file.param' has incorrect format.")
                                    glib.printErrorMessage("Format is key=value. Parameter given is: '%s'." % item)
                                    return 1
                
                # associatedServices list => optional
                serviceAssociated = []
                if instanceParams.has_key('service-discovery.file.associatedService'):
                    if len(instanceParams['service-discovery.file.associatedService']) > 0:
                        for item in instanceParams['service-discovery.file.associatedService']:
                            if item != "":
                                serviceAssociated.append(item) 
 
                # add / replace entry
                if self.serviceNameExists(serviceName):
                    errorcode = self.addService(serviceName, serviceEndpoint, serviceType, serviceVersion, serviceSite, 
                                                serviceVos, serviceParams, serviceWsdl, serviceAdministration, serviceAssociated)
                else:                        
                    errorcode = self.replaceService(serviceName, serviceEndpoint, serviceType, serviceVersion, serviceSite, 
                                                    serviceVos, serviceParams, serviceWsdl, serviceAdministration, serviceAssociated)
                                    
                if errorcode:
                    print "Configuring file based Service Discovery                  ",
                    glib.printFailedMessage()
                    return 1                    
        
        # write file
        errorcode = self.writeFile()
        if errorcode:
            print "Configuring file based Service Discovery                  ",
            glib.printFailedMessage()
    
        if self.verbose:
            print "Configuring file based Service Discovery                  ",
            glib.printOkMessage()
        else:
            glib.printOkMessage()
        
        return 0
    

    #################################################################
    # serviceNameExists
    #
    # does service name exist
    #
    # params serviceName serviceName
    #
    # return 0 service name exists
    #     1 service name does not exist
    #     
    #################################################################
    def serviceNameExists(self, serviceName):

        nodes = xpath.Evaluate("/services/service[@name='%s']" % serviceName, self.doc.documentElement)
        if len(nodes) > 0:
            return 0
        else:
            return 1


    #################################################################
    # addService
    #
    # add service to service list
    #
    # params serviceName             name of service
    #        serviceEndpoint         endpoint for service
    #        serviceType             type of service
    #        serviceVersion          version of service
    #        serviceSite             site for service
    #        serviceVos              array of vos for service
    #        serviceParams           array of service extra parameters
    #        serviceWsdl             service Wsdl
    #        serviceAdministration   service administration
    #        serviceAssociated       array of associated services
    #
    # return 0 ok 
    #        1 error
    #     
    #################################################################  
    def addService(self, serviceName, serviceEndpoint, serviceType, serviceVersion, 
                    serviceSite, serviceVos, serviceParams, serviceWsdl, serviceAdministration, 
                    serviceAssociated):
        
        try:
            if self.verbose:
                print "Creating service '%s' " % serviceName,
            
            #------------------------------------------------------------
            # Create individual elements
            #------------------------------------------------------------
        
            # service element
            serviceElement = self.doc.createElement("service")
            serviceElement.setAttribute("name",serviceName)
        
            # parameters element
            parametersElement = self.doc.createElement("parameters")
            
            # endpoint element
            endpointElement = self.doc.createElement("endpoint")
            endpointElementText = self.doc.createTextNode(serviceEndpoint)
            endpointElement.appendChild(endpointElementText)
            parametersElement.appendChild(endpointElement)
        
            # type element
            typeElement = self.doc.createElement("type")
            typeElementText = self.doc.createTextNode(serviceType)
            typeElement.appendChild(typeElementText)
            parametersElement.appendChild(typeElement)
        
            # version element
            versionElement = self.doc.createElement("version")
            versionElementText = self.doc.createTextNode(serviceVersion)
            versionElement.appendChild(versionElementText)
            parametersElement.appendChild(versionElement)
        
            # site element (optional)
            if serviceSite != "":
                siteElement = self.doc.createElement("site")
                siteElementText = self.doc.createTextNode(serviceSite)
                siteElement.appendChild(siteElementText)
                parametersElement.appendChild(siteElement)
        
            # wsdl element (optional)
            if serviceWsdl != "":
                wsdlElement = self.doc.createElement("wsdl")
                wsdlElementText = self.doc.createTextNode(serviceWsdl)
                wsdlElement.appendChild(wsdlElementText)
                parametersElement.appendChild(wsdlElement)
         
            # administration element (optional)
            if serviceAdministration != "":
                administrationElement = self.doc.createElement("administration")
                administrationElementText = self.doc.createTextNode(serviceAdministration)
                administrationElement.appendChild(administrationElementText)
                parametersElement.appendChild(administrationElement)
                        
            # vo list (optional)
            if len(serviceVos) != 0:
                
                volistElement = self.doc.createElement("volist")

                for vo in serviceVos:
                                        
                    # create element
                    voElement = self.doc.createElement("vo")
                    voElementText = self.doc.createTextNode(vo)
                    voElement.appendChild(voElementText)
                    
                    # add element
                    volistElement.appendChild(voElement)

                parametersElement.appendChild(volistElement)

            # param list (optional)
            if len(serviceParams) != 0:
                
                for parameter in serviceParams:
                
                    param = parameter.split("=")
                    paramName  = param[0]
                    paramValue = param[1]
                    
                    paramElement = self.doc.createElement("param")
                    paramElement.setAttribute("name",paramName)
                    paramElementText = self.doc.createTextNode(paramValue)
                    paramElement.appendChild(paramElementText)
                                    
                    parametersElement.appendChild(paramElement)            
      
            #------------------------------------------------------------
            # Put /service/parameter tree together
            #------------------------------------------------------------
            serviceElement.appendChild(parametersElement)

            #------------------------------------------------------------
            # associated services (optional)
            #------------------------------------------------------------            
            if len(serviceAssociated) != 0:
                
                associatedListElement = self.doc.createElement("associatedservices")

                for name in serviceAssociated:
                                        
                    # create element
                    associatedElement = self.doc.createElement("name")
                    associatedElementText = self.doc.createTextNode(name)
                    associatedElement.appendChild(associatedElementText)
                    
                    # add element
                    associatedListElement.appendChild(associatedElement)

                serviceElement.appendChild(associatedListElement)
        
            # combine root element
            self.doc.documentElement.appendChild(serviceElement)
            
        except ValueError: 
            if not self.verbose:
                print "Creating service '%s' " % serviceName,
            self.glib.printFailedMessage()
            return 1

        if self.verbose:
            self.glib.printOkMessage()
        return 0


    #################################################################
    # removeService
    #
    # remove service in service list
    #
    # params serviceName      name of service
    #
    # return 0 ok 
    #     1 error
    #     
    #################################################################  
    def removeService(self, serviceName):
        
        try:    
            nodes = xpath.Evaluate("/services/service[@name='%s']" % serviceName, self.doc.documentElement)
    
            if len(nodes) > 0:
                for node in nodes:                
                    self.doc.documentElement.removeChild(node)
                    if self.verbose:
                        print "Removing service '%s' " % serviceName,
                        self.glib.printOkMessage()
                return 0            
            else:
                print "Removing service '%s' " % serviceName,
                self.glib.printFailedMessage()
                self.glib.printWarningMessage("Service '%s' does not exist in service list." % serviceName)
                self.glib.printWarningMessage("Service '%s' not removed from service list." % serviceName)
                return 1
        except ValueError, ex:
            print "Removing service '%s' " % serviceName,
            self.glib.printFailedMessage()
            self.glib.printErrorMessage("Errror removing old service entry '%s':" % serviceName)
            raise ex
        
        
        
    #################################################################
    # replaceService
    #
    # replace service in service list
    #
    # params serviceName             name of service
    #        serviceEndpoint         endpoint for service
    #        serviceType             type of service
    #        serviceVersion          version of service
    #        serviceSite             site for service
    #        serviceVos              array of vos for service
    #        serviceParams           array of service extra parameters
    #        serviceWsdl             service Wsdl
    #        serviceAdministration   service administration
    #        serviceAssociated       array of associated services
    #    
    # return 0 ok 
    #        1 error
    #     
    #################################################################  
    def replaceService(self, serviceName, serviceEndpoint, serviceType, serviceVersion, serviceSite, 
                        serviceVos, serviceParams, serviceWsdl, serviceAdministration, 
                        serviceAssociated):                      
            
        errorcode  = self.removeService(serviceName)
        errorcode += self.addService(serviceName, serviceEndpoint, serviceType, serviceVersion, 
                                     serviceSite, serviceVos, serviceParams, serviceWsdl, serviceAdministration, 
                                     serviceAssociated)
            
        if errorcode:
            print "Replacing service entry '%s' " % serviceName,
            self.glib.printFailedMessage()
            return 1
        
        if self.verbose:
            print "Replacing service entry '%s' " % serviceName,
            self.glib.printOkMessage()
        return 0
        

    #################################################################
    # writeFile
    #
    # write file
    #
    # return 0 ok 
    #     1 error
    #     
    #################################################################        
    def writeFile(self):
        
        if self.verbose:
            print "Writing service discovery file                            ",
        try:
            file = open("%s.1" % self.serviceDiscoveryFile, "w")
            PrettyPrint(self.doc.documentElement, stream=file)
            file.close()

            # cleanup			
            os.rename('%s.1' % self.serviceDiscoveryFile, '%s' % self.serviceDiscoveryFile)

        except IOError, ex:
            if not self.verbose:
                print "Writing service discovery file                            ",
            self.glib.printFailedMessage()
            raise ex
            return 1
        
        if self.verbose:
            self.glib.printOkMessage()
        return 0
            
